/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qformatscheme.h"

/*!
	\file qformatscheme.cpp
	\brief Implementation of QFormatScheme
	
	\see QFormatScheme
*/

/*!
	\ingroup language
	@{
	
	\class QFormatScheme
	\brief A storage/configuration class for shared highlighting formats
	
	It stores text formats used by highlighters interfaces and provides
	a default serializing format in QXF format (XML-based).
	
	\see QLanguageFactory
	\see QHighlighter
*/

#include "qformat.h"

#include <QFile>
#include <QDomText>
#include <QSettings>
#include <QStringList>
#include <QDomElement>
#include <QDomDocument>

#define QFORMAT_VERSION "1.0"

static bool bool_cast(const QString& s)
{
	return !QString::compare(s, QLatin1String("true")) || s.toUInt() == 1;
}

QVector<QString>& QFormatScheme::m_globalFormatKeys()
{
	static QVector<QString> _d;
	return _d;
}

QVector<QFormat>& QFormatScheme::m_globalFormatValues()
{
	static QVector<QFormat> _d;
	return _d;
}

QHash<QString, int>& QFormatScheme::m_globalFormatIds()
{
	static QHash<QString, int> _d;
	return _d;
}

/*!
	\brief Constructor
*/
QFormatScheme::QFormatScheme(QObject *p)
 : QObject(p)
{
	if ( m_globalFormatKeys().isEmpty() )
	{
		m_globalFormatKeys() << "normal";
		m_globalFormatValues() << QFormat();
		m_globalFormatIds()["normal"] = 0;
	}
}

/*!
	\brief Constructor
	\param f Filename of a format settings file to load
*/
QFormatScheme::QFormatScheme(const QString& f, QObject *p)
 : QObject(p)
{
	if ( m_globalFormatKeys().isEmpty() )
	{
		m_globalFormatKeys() << "normal";
		m_globalFormatValues() << QFormat();
		m_globalFormatIds()["normal"] = 0;
	}
	
	load(f);
}

/*!
	\brief Destructor
*/
QFormatScheme::~QFormatScheme()
{
}

/*!
	\brief Re-initialize the format scheme
	
	Calling this method leaves the format scheme with only one
	format : the "normal" one, set to a default-constructed QFormat
*/
void QFormatScheme::clear()
{
	m_formats.clear();
}

/*!
	\brief Load format settings from a file
	\param f file to load data from

	The default implementation loads data in QXF format (XML-based)

	\note Previous content is discarded
*/
void QFormatScheme::load(const QString& f)
{
	clear();
	m_settings = f;

	QFile settings(f);
	
	if ( settings.open(QFile::ReadOnly | QFile::Text) )
	{
		QDomDocument doc;
		doc.setContent(&settings);
		
		load(doc.documentElement());
	}
}

/*!
	\brief Save the format settings to a file
	\param f target file (if none specified, last value passed to load is used)

	The default implementation saves data in QXF format (XML-based)
*/
void QFormatScheme::save(const QString& f) const
{
	QFile settings(f.count() ? f : m_settings);
	
	if ( settings.open(QFile::WriteOnly | QFile::Text) )
	{
		QDomDocument doc("QXF");
		QDomElement root = doc.createElement("QXF");
		save(root);
		doc.appendChild(root);
		settings.write(doc.toByteArray(4));
	}
}

/*!
	\overload
	\param elem Source element to scan
	\param ignoreNewIds whether unknown format identifiers should be ignored
	
	The given dom element must contain a proper version attribute and format
	data as child elements (&lt;format&gt; tags)
	
	\note Previous content is not discarded
*/
void QFormatScheme::load(const QDomElement& elem, bool ignoreNewIds)
{
	if ( elem.attribute("version") < QFORMAT_VERSION )
	{
		qWarning("Format encoding version mismatch : [found]%s != [expected]%s",
				qPrintable(elem.attribute("version")),
				QFORMAT_VERSION);
		
		return;
	}
	
	QDomElement e, c;
	QDomNodeList l, f = elem.elementsByTagName("format");
	
	for ( int i = 0; i < f.count(); ++i )
	{
		e = f.at(i).toElement();
		
		int idx = m_globalFormatIds().value(e.attribute("id"), -1);
		
		if ( ignoreNewIds && (idx != -1) && m_formats.contains(idx) )
			continue;
		
		l = e.childNodes();
		
		QFormat fmt;
		
		for ( int j = 0; j < l.count(); ++j )
		{
			c = l.at(j).toElement();
			
			if ( c.isNull() )
				continue;
			
			QString field = c.tagName(),
					value = c.firstChild().toText().data();
			
			if ( field == "bold" )
				fmt.weight = bool_cast(value) ? QFont::Bold : QFont::Normal;
			else if ( field == "italic" )
				fmt.italic = bool_cast(value);
			else if ( field == "overline" )
				fmt.overline = bool_cast(value);
			else if ( field == "underline" )
				fmt.underline = bool_cast(value);
			else if ( field == "strikeout" )
				fmt.strikeout = bool_cast(value);
			else if ( field == "waveUnderline" )
				fmt.waveUnderline = bool_cast(value);
			else if ( field == "color" || field == "foreground" )
				fmt.foreground = QColor(value);
			else if ( field == "background" )
				fmt.background = QColor(value);
			else if ( field == "linescolor" )
				fmt.linescolor = QColor(value);
			
		}
		
		setFormat(e.attribute("id"), fmt);
	}
}

/*!
	\overload
*/
void QFormatScheme::save(QDomElement& elem) const
{
	QDomDocument doc = elem.ownerDocument();
	elem.setAttribute("version", QFORMAT_VERSION);
	
	QStringList fl = formats();
	
	foreach ( QString fid, fl )
	{
		QDomText t;
		QDomElement f, c = doc.createElement("format");
		
		c.setAttribute("id", fid);
		
		const QFormat& fmt = format(fid);
		
		f = doc.createElement("bold");
		t = doc.createTextNode((fmt.weight == QFont::Bold) ? "true" : "false");
		f.appendChild(t);
		c.appendChild(f);
		
		f = doc.createElement("italic");
		t = doc.createTextNode(fmt.italic ? "true" : "false");
		f.appendChild(t);
		c.appendChild(f);
		
		f = doc.createElement("overline");
		t = doc.createTextNode(fmt.overline ? "true" : "false");
		f.appendChild(t);
		c.appendChild(f);
		
		f = doc.createElement("underline");
		t = doc.createTextNode(fmt.underline ? "true" : "false");
		f.appendChild(t);
		c.appendChild(f);
		
		f = doc.createElement("strikeout");
		t = doc.createTextNode(fmt.strikeout ? "true" : "false");
		f.appendChild(t);
		c.appendChild(f);
		
		f = doc.createElement("waveUnderline");
		t = doc.createTextNode(fmt.waveUnderline ? "true" : "false");
		f.appendChild(t);
		c.appendChild(f);
		
		if ( fmt.foreground.isValid() )
		{
			f = doc.createElement("foreground");
			t = doc.createTextNode(fmt.foreground.name());
			f.appendChild(t);
			c.appendChild(f);
		}
		
		if ( fmt.background.isValid() )
		{
			f = doc.createElement("background");
			t = doc.createTextNode(fmt.background.name());
			f.appendChild(t);
			c.appendChild(f);
		}
		
		if ( fmt.linescolor.isValid() )
		{
			f = doc.createElement("linescolor");
			t = doc.createTextNode(fmt.linescolor.name());
			f.appendChild(t);
			c.appendChild(f);
		}
		
		elem.appendChild(c);
	}
}

/*!
	\overload
	\brief Load format data from a QSettings object
	\param s QSettings object from which data will be fetched
	\param ignoreNewIds whether unknown format identifiers should be ignored
	
	The QSettings object is assumed to be initialized properly and to
	point to a correct location.
	
	\note Previous content is not discarded
*/
void QFormatScheme::load(QSettings& s, bool ignoreNewIds)
{
	QString version = s.value("version").toString();
	
	if ( version < QFORMAT_VERSION )
	{
		qWarning("Format encoding version mismatch : [found]%s != [expected]%s",
				qPrintable(version),
				QFORMAT_VERSION);
		
		return;
	}
	
	s.beginGroup("data");
	
	QStringList l = s.childGroups();
	
	foreach ( QString id, l )
	{
		int idx = m_globalFormatIds().value(id, -1);
		
		if ( ignoreNewIds && (idx != -1) && m_formats.contains(idx) )
			continue;
		
		s.beginGroup(id);
		
		QFormat fmt;
		QStringList fields = s.childKeys();
		
		foreach ( QString field, fields )
		{
			QString value = s.value(field).toString();
			
			if ( field == "bold" )
				fmt.weight = bool_cast(value) ? QFont::Bold : QFont::Normal;
			else if ( field == "italic" )
				fmt.italic = bool_cast(value);
			else if ( field == "overline" )
				fmt.overline = bool_cast(value);
			else if ( field == "underline" )
				fmt.underline = bool_cast(value);
			else if ( field == "strikeout" )
				fmt.strikeout = bool_cast(value);
			else if ( field == "waveUnderline" )
				fmt.waveUnderline = bool_cast(value);
			else if ( field == "color" || field == "foreground" )
				fmt.foreground = QColor(value);
			else if ( field == "background" )
				fmt.background = QColor(value);
			else if ( field == "linescolor" )
				fmt.linescolor = QColor(value);
			
		}
		
		setFormat(id, fmt);
		s.endGroup();
	}
	
	s.endGroup();
}

/*!
	\overload
*/
void QFormatScheme::save(QSettings& s) const
{
	s.setValue("version", QFORMAT_VERSION);
	
	s.beginGroup("data");
	
	QStringList fl = formats();
	
	foreach ( QString fid, fl )
	{
		s.beginGroup(fid);
		
		const QFormat& fmt = format(fid);
		
		s.setValue("bold", (fmt.weight == QFont::Bold) ? "true" : "false");
		s.setValue("italic", fmt.italic ? "true" : "false");
		s.setValue("overline", fmt.overline ? "true" : "false");
		s.setValue("underline", fmt.underline ? "true" : "false");
		s.setValue("strikeout", fmt.strikeout ? "true" : "false");
		s.setValue("waveUnderline", fmt.waveUnderline ? "true" : "false");
		
		if ( fmt.foreground.isValid() )
		{
			s.setValue("foreground", fmt.foreground.name());
		}
		
		if ( fmt.background.isValid() )
		{
			s.setValue("background", fmt.background.name());
		}
		
		if ( fmt.linescolor.isValid() )
		{
			s.setValue("linescolor", fmt.linescolor.name());
		}
		
		s.endGroup();
	}
	
	s.endGroup();
}

/*!
	\return A list of available (local) format keys
*/
QStringList QFormatScheme::formats() const
{
	QStringList l;
	QList<int> lid = m_formats.keys();
	
	foreach ( int id, lid )
		l << m_globalFormatKeys().at(id);
	
	return l;
}

/*!
	\return The format key associated to integer format id \a ifid
*/
QString QFormatScheme::id(int ifid) const
{
	return m_globalFormatKeys().at(ifid);
}

/*!
	\return The integer format id associated to format key \a fid
*/
int QFormatScheme::id(const QString& sfid) const
{
	return m_globalFormatIds().value(sfid, -1);
}

/*!
	\return The text format associated with format key \a fid
	
	\warning Use at your own risks : if there are no format associated
	with the requested id this function may crash or lead to inconsitencies
*/
QFormat& QFormatScheme::formatRef(int ifid)
{
	return m_formats.contains(ifid) ? m_formats[ifid] : m_globalFormatValues()[ifid];
}

/*!
	\return The a reference to the text format associated with format key \a fid
	
	\warning Use at your own risks : if there are no format associated
	with the requested id this function may crash or lead to inconsitencies
*/
QFormat& QFormatScheme::formatRef(const QString& sfid)
{
	return m_formats[id(sfid)];
}

/*!
	\return The text format associated with format key \a fid
*/
QFormat QFormatScheme::format(int ifid) const
{
	QFormat f = m_formats.value(ifid, ifid < m_globalFormatValues().count() ? m_globalFormatValues().at(ifid) : QFormat());
	
	//qDebug("%i -> %s", ifid, qPrintable(f.foreground.name()));
	
	return f;
}

/*!
	\return The text format associated with format key \a fid
*/
QFormat QFormatScheme::format(const QString& sfid) const
{
	return format(id(sfid));
}

/*!
	\brief Set text format for key
	\param fid Format key
	\param fmt Format value
*/
void QFormatScheme::setFormat(const QString& fid, const QFormat& fmt, bool overrideGlobal)
{
	const int idx = m_globalFormatIds().value(fid, -1);
	
	if ( idx != -1 )
	{
		//qDebug("localizing format %s [%i]", qPrintable(fid), idx);
		m_formats[idx] = fmt;
		
		if ( overrideGlobal )
			m_globalFormatValues()[idx] = fmt;
	} else {
		//qDebug("adding format %s [%i] -> %s", qPrintable(fid), m_globalFormatKeys().count(), qPrintable(fmt.foreground.name()));
		
		int id = m_globalFormatKeys().count();
		
		m_globalFormatKeys() << fid;
		m_globalFormatValues() << fmt;
		m_globalFormatIds()[fid] = id;
		
		m_formats[id] = fmt;
	}
}

/*! @} */
