/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qlanguagedefinition.h"

/*!
	\file qlanguagedefinition.cpp
	\brief Implementation of QLanguageDefinition
	
	\see QLanguageDefinition
*/

/*!
	\ingroup language
	@{
	
	\class QLanguageDefinition
	\brief Interface for language definition.
	
	This class is meant to be subclassed, see \see QGenericDefinition for more
	informations, and added to a QLanguageFactory.
	
	A language definition is a wrapper that creates interfaces for a given file
	extension from internally handled data (XML files in the case of 
	QGenericDefinition)
	
	\see QLanguageFactory
*/

#include "qdocument.h"
#include "qdocument_p.h"
#include "qdocumentline.h"
#include "qdocumentcursor.h"
#include "qdocumentbuffer.h"

#include "qlanguagefactory.h"

#include <QKeyEvent>

int QLanguageDefinition::m_contextCount = 0;
QVector<QLanguageDefinition*> QLanguageDefinition::m_definitions;

/*!
	The returned id does not have the Language flag set (context id packing)
*/
int QLanguageDefinition::id(const QString& lang)
{
	for ( int i = 0; i < m_definitions.count(); ++i )
	{
		QLanguageDefinition *l = m_definitions.at(i);
		
		if ( !l )
			continue;
		
		if ( l->language() == lang )
			return i;
	}
	
	return -1;
}

/*!
	The passed id may or may not have the Language flag (context id packing)
*/
QLanguageDefinition* QLanguageDefinition::instance(int id)
{
	id &= ~(Language);
	
	return (id >= 0 && id < m_definitions.count()) ? m_definitions.at(id) : 0;
}

/*!
	\brief Empty constructor
*/
QLanguageDefinition::QLanguageDefinition()
 : m_id(m_definitions.count())
{
	m_definitions << this;
}

/*!
	\brief Empty destructor
*/
QLanguageDefinition::~QLanguageDefinition()
{
	m_definitions[m_id] = 0;
}

/*!
	\return The language id of this definition
*/
int QLanguageDefinition::languageId() const
{
	return m_id;
}

/*!
	\brief context id to name
*/
QString QLanguageDefinition::contextName(int id) const
{
	int idx = m_contextIds.indexOf(id);
	
	return m_contextNames.at(idx);
}

/*!
	\brief context name to id
*/
int QLanguageDefinition::contextId(const QString& s) const
{
	int idx = m_contextNames.indexOf(s);
	
	return m_contextIds.at(idx);
}

/*!
	\brief register a context
	
	Meant to be used by subclasses upon construction
*/
int QLanguageDefinition::addContext(const QString& cxt)
{
	m_contextIds << m_contextCount;
	m_contextNames << cxt;
	++m_contextCount;
	return m_contextCount - 1;
}

/*!
	\fn QLanguageDefinition::language() 
	
	\return The language supported by this definition
*/

/*!
	\fn QLanguageDefinition::extensions()
	
	\return the file extensions corresponding to the supported language
	
	\see language()
	\see QFileInfo::completeSuffix()
*/

int QLanguageDefinition::tokenFormat(int tokenId) const
{
	return tokenId & TokenIdMask;
}

int QLanguageDefinition::contextFormat(int contextId) const
{
	return contextId & ContextIdMask;
}

/*!
	\brief Entry point for syntax highlighting
*/
int QLanguageDefinition::tokenize(QDocument *d, int line, int count)
{
	int m_line = line, m_column = 0;
	QDocumentBuffer::iterator it = d->impl()->find(line);
	
	return qMax(tokenize(it, m_line, m_column, line + count, 0, 0), count);
}

/*!
	\brief entry point for nested syntax highlighting
	\return the number of lines actually tokenized
	
	This entry point may be called by other definitions to embed a language in another.
	
	In this case the \a depth indicates the level of tokenize() recursion. When the \a depth is
	null it means the call originates from the definition itself as the other tokenize()
	method has been called by the document.
	
	When the \a depth is non null tokenization must be ATTEMPTED unless \a cxt is a valid context
	index for the current line in which case the embedding has already happened and an update is
	requested for an effectively embedded region. Otherwise, if the first character does not lead
	to a match then nothing should be done and in particular the \a line and \a column should not
	be altered.
	
*/
int QLanguageDefinition::tokenize(QDocumentBuffer::iterator& it, int& line, int& column, int extent, int depth, int cxt)
{
	Q_UNUSED(it)
	Q_UNUSED(line)
	Q_UNUSED(column)
	Q_UNUSED(extent)
	Q_UNUSED(depth)
	Q_UNUSED(cxt)
	
	return 0;
}

/*!
	\brief Return the string starting a single line comment, if any offered by the language
*/
QString QLanguageDefinition::singleLineComment() const
{
	return QString();
}

/*!
	\brief Let language specify which line mark should be toggled by left clicking a line mark panel
*/
QString QLanguageDefinition::defaultLineMark() const
{
	return QString();
}

/*!
	\brief Brace matching entry point
*/
void QLanguageDefinition::clearMatches(QDocument *d)
{
	Q_UNUSED(d)
}

/*!
	\brief Brace matching entry point
*/
void QLanguageDefinition::match(QDocumentCursor& c)
{
	Q_UNUSED(c)
}

/*!
	\brief Return the indent to use when inserting a line at a given cursor position
*/
QString QLanguageDefinition::indent(const QDocumentCursor& c)
{
	Q_UNUSED(c)
	
	return QString();
}

/*!
	\brief Determines whether the given key event at the given position should cause unindent to happen
*/
bool QLanguageDefinition::unindent (const QDocumentCursor& c, const QString& ktxt)
{
	Q_UNUSED(c)
	Q_UNUSED(ktxt)
	
	return false;
}

/*!
	\brief Expand a collapsed block at a given line
*/
void QLanguageDefinition::expand(QDocument *d, int line)
{
	Q_UNUSED(d)
	Q_UNUSED(line)
}

/*!
	\brief Collapse a text block at a given line
*/
void QLanguageDefinition::collapse(QDocument *d, int line)
{
	Q_UNUSED(d)
	Q_UNUSED(line)
}

/*!
	\brief Compute the collapse state of a line
*/
int QLanguageDefinition::blockFlags(QDocument *d, int line, int depth) const
{
	Q_UNUSED(d)
	Q_UNUSED(line)
	Q_UNUSED(depth)
	
	return 0;
}

/*!
	\brief Compares two parentheses
	
	The default implementation simply compares the format fields of each token id
*/
bool QLanguageDefinition::parenMatch(int tokenId1, const QStringRef prod1, int tokenId2, const QStringRef prod2) const
{
	Q_UNUSED(prod1)
	Q_UNUSED(prod2)
	
	return (tokenId1 & TokenIdMask) == (tokenId2 & TokenIdMask);
}

/*! @} */
